"""
This module draws diagrams of the state machines.

Usage:
    ~/ska-src/ska-tango-base$ docker run --rm -ti -v $PWD:/app continuumio/miniconda3 bash

    (base) root@293f3b699c9b:#
    $ conda install --yes pygraphviz
    $ pip install transitions
    $ apt-get update && apt-get install gsfonts
    $ cd /app/docs/source
    $ python draw_state_machines.py

(Or see the top level Makefile)

"""
import importlib
import sys

from unittest import mock
from transitions.extensions import GraphMachine

import ska_tango_base


def patch_in_graph_machine():
    module_names = [
        "ska_tango_base.base.op_state_model",
        "ska_tango_base.base.admin_mode_model",
        "ska_tango_base.subarray.subarray_obs_state_model",
        "ska_tango_base.csp.csp_subelement_obs_state_model",
    ]
    with mock.patch("transitions.extensions.LockedMachine", GraphMachine):
        for module_name in module_names:
            mod = sys.modules[module_name]
            importlib.reload(mod)


def draw(machine_class, target_filename):
    machine_name = machine_class.__name__
    graphing_options = {
        "title": f"\n{machine_name}",
        "show_conditions": True,
        "show_state_attributes": True,
        "show_auto_transitions": False,
    }
    machine = machine_class(None, **graphing_options)
    machine.get_graph().draw(target_filename, prog='dot')


if __name__ == "__main__":
    patch_in_graph_machine()
    draw(
        ska_tango_base.base.op_state_model._OpStateMachine,
        "api/base/_OpStateMachine_autogenerated.png",
    )
    draw(
        ska_tango_base.base.admin_mode_model._AdminModeMachine,
        "api/base/_AdminModeMachine_autogenerated.png",
    )
    draw(
        ska_tango_base.subarray.subarray_obs_state_model._SubarrayObsStateMachine,
        "api/subarray/_SubarrayObsStateMachine_autogenerated.png",
    )
    draw(
        ska_tango_base.csp.csp_subelement_obs_state_model._CspSubElementObsStateMachine,
        "api/csp/_CspSubElementObsStateMachine_autogenerated.png",
    )
